rules = {
    "noun_num_quant" : [ # more than four times, about 127 gigatonnes, up to 400 people
        {
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "amod", "compound", "nummod", "nmod", "advmod", "npadvmod", "dobj"]}, "POS": "NUM"}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">>",
            "RIGHT_ID": "adv",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "amod", "compound", "nummod", "nmod", "advmod", "npadvmod"]}, "POS": {"NOT_IN": ["NUM"]}}
        },
    ],

    "noun_num_quant_2" : [ # higher/lower than 8.90 ksi
        {
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "amod", "compound", "nummod", "nmod", "advmod", "npadvmod"]}, "POS": "NUM"}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": "<<",
            "RIGHT_ID": "adj",
            "RIGHT_ATTRS": {"DEP": {"IN": ["acomp", "amod","punct"]}, "POS": "ADJ"}
        }
    ],

    "noun_num_quant_3" : [ # higher than 130
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "amod", "compound", "nummod", "nmod", "advmod", "npadvmod", "pobj"]}, "POS": "NUM"}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": "<<",
            "RIGHT_ID": "adj",
            "RIGHT_ATTRS": {"DEP": {"IN": ["acomp", "amod"]}, "POS": "ADJ", "ORTH": {"IN": ["higher", "lower"]}}
        }
    ],

    "noun_num_quant_4" : [ # ~80%
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "amod", "compound", "nummod", "nmod", "advmod", "npadvmod", "pobj"]}, "POS": "NUM"}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ";",
            "RIGHT_ID": "sign",
            "RIGHT_ATTRS": {"DEP": {"IN": ["punct", "attr"]}, "POS": {"IN": ["PUNCT", "X", "SYM"]}, "ORTH": {"IN": ["~","=","<",">", "≈"]}}
        }
    ],

    "noun_num_right_noun" : [
        {
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"DEP": "nummod", "POS": "NUM"}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "right_noun",
            "RIGHT_ATTRS": {"DEP": {"IN": ["pobj"]}, "POS": {"IN": ["PROPN", "NOUN"]}}
        },
    ],

    "noun_num_adp_right_noun" : [ # 22 miles per hour, 13% of Aeronautics, 127 gigatonnes of mass, 60% of workers
        {
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"DEP": "nummod", "POS": "NUM"}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "adp",
            "RIGHT_ATTRS": {"DEP": {"IN": ["prep", "appos"]}, "POS": {"IN": ["ADP"]}, "ORTH": {"IN": ["of", "per"]}}
        },
        {
            "LEFT_ID": "adp",
            "REL_OP": ">",
            "RIGHT_ID": "right_noun",
            "RIGHT_ATTRS": {"DEP": {"IN": ["pobj", "appos"]}, "POS": {"IN": ["PROPN", "NOUN", "ADJ"]}}
        },
    ],

    "num_num_adp_right_noun" : [ # One fourth of population
        {
            "RIGHT_ID": "number1",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        },
        {
            "LEFT_ID": "number1",
            "REL_OP": ">",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"DEP": "nummod", "POS": "NUM"}
        },
        {
            "LEFT_ID": "number1",
            "REL_OP": ">",
            "RIGHT_ID": "adp",
            "RIGHT_ATTRS": {"DEP": {"IN": ["prep", "appos"]}, "POS": {"IN": ["ADP"]}, "ORTH": {"IN": ["of", "per"]}}
        },
        {
            "LEFT_ID": "adp",
            "REL_OP": ">",
            "RIGHT_ID": "right_noun",
            "RIGHT_ATTRS": {"DEP": {"IN": ["pobj", "appos"]}, "POS": {"IN": ["PROPN", "NOUN"]}}
        },
    ],


    "num_symbol" : [ # $62, $ billion, US$ million
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM"}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "symbol",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "nmod"]}, "POS": "SYM", "ORTH": {"NOT_IN": ["#"]}}
        },
    ],

    "symbol_num" : [
        {
            "RIGHT_ID": "symbol",
            "RIGHT_ATTRS": {"POS": "SYM"}
        },
        {
            "LEFT_ID": "symbol",
            "REL_OP": ";",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM"}
        },
    ],

    "noun_num" : [ # 0.4 percent, 127 gigatonnes, 2.1%, six years, million contract (from $48.2 million army contract)
        {
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN", "INTJ", "ADJ"]}, "ORTH": {"NOT_IN": ["#", "ers"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "nmod", "nummod", "compound", "amod", "nsubj"]}, "POS": "NUM"}
        },
    ],

    "num_noun" : [ # four times, 6.5 times, 20 percent
        {
            "RIGHT_ID": "num",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "nmod", "nummod", "attr"]}, "POS": "NUM"}
        },
        {
            "LEFT_ID": "num",
            "REL_OP": ">",
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN"]}}
        },
    ],

    "num_direct_propn" : [ # 2.4 Ghz, 30 Mbps, 60 GB, 450 Nm, 0.1 m2, 50 mL
        # 575 Willbraham, 8 AB8, 30 TFSI
        {
            "RIGHT_ID": "num",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "nmod", "nummod"]}, "POS": "NUM"}
        },
        {
            "LEFT_ID": "num",
            "REL_OP": ".",
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN"]}}
        },
    ],

    "lonely_num" : [ # just the number: 21, thirteen, million
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM"}
        }
    ],

    "noun_compound_num" : [ # 15 year-old students
        {
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "compound",
            "RIGHT_ATTRS": {"DEP": "compound", "POS": {"IN": ["ADJ"]}}
        },
        {
            "LEFT_ID": "compound",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "amod", "nummod", "nmod"]}, "POS": "NUM"}
        },
    ],

    "noun_adj_num" : [ # double 120 decibels, 600 people worldwide, 6kW electric motor, million Australian dollars, 14 first-team players
        # billion hostile takeover, million year-old bedrock
        {
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "amod", "compound", "nummod", "nmod", "advmod", "npadvmod"]}, "POS": {"IN": ["NUM"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "adv",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "amod", "compound", "nummod", "nmod", "advmod", "npadvmod"]},
                            "POS": {"IN": ["ADJ", "SCONJ", "ADV", "PART", "ADP", "DET", "SYM"]}}
        },
    ],

    "adj_num" : [ # 300,000 year-old, 5 kW
        {
            "RIGHT_ID": "adv",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "amod", "nummod", "nmod"]}, "POS": {"IN": ["ADJ"]}}
        },
        {
            "LEFT_ID": "adv",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "amod", "compound", "nummod", "nmod", "advmod", "npadvmod"]}, "POS": "NUM"}

        },
    ],

    "adj_noun_num" : [ # 30 years old, ten months old, 90 day old
        {
            "RIGHT_ID": "adj",
            "RIGHT_ATTRS": {"DEP": {"IN": ["acomp", "amod"]}, "POS": {"IN": ["ADJ"]}, "ORTH": {"IN": ["old"]}}
        },
        {
            "LEFT_ID": "adj",
            "REL_OP": ">",
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"DEP": {"IN": ["npadvmod"]}, "POS": {"IN": ["NOUN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "nmod", "nummod", "compound"]}, "POS": "NUM"}
        }
    ],
    "num_direct_noun" : [ # 87 yo, 50 year-old, 50 mph, 8.5 kW, 0.6 percent/%, one league
        {
            "RIGHT_ID": "num",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "nmod", "nummod"]}, "POS": "NUM"}
        },
        {
            "LEFT_ID": "num",
            "REL_OP": ".",
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"DEP": {"IN": ["compound"]}, "POS": {"IN": ["NOUN"]}}
        },
    ],

    "num_quantmod" : [ # around/about 600, just/only 799, over/under thousand, more than four times*, up to 400*
        # also between 62 (between $62 and $68)
        # * as many as 14 (14,as;14,many;14,as -> as separate matches)
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM"}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "quant",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "amod", "compound", "nummod", "nmod", "advmod", "npadvmod"]},
                            "POS": {"IN": ["ADJ", "SCONJ", "ADV", "PART", "ADP", "DET"]}}
        },
    ],

    "minus_num" : [ # minus 0.1
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM"}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": "<",
            "RIGHT_ID": "quant",
            "RIGHT_ATTRS": {"ORTH": "minus"}
        },
    ],

    "minus_num_2" : [ # minus 1.7%
        {
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": "NOUN"}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "quant",
            "RIGHT_ATTRS": {"ORTH": "minus"}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM"}
        },
    ],

    "quantmod_direct_num" : [ # averaging 15.9, rose/fell 0.6%, lost/gained 286 gigatonnes
        # close one, was five, totalling twenty, plummeted 190, climbed 0.1%
        # down 0.03%
        {
            "RIGHT_ID": "num",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        },
        {
            "LEFT_ID": "num",
            "REL_OP": ";",
            "RIGHT_ID": "quant",
            "RIGHT_ATTRS": {"DEP": {"IN": ["advcl", "advmod", "amod", "conj", "acl", "prep", "prt", "ROOT", "pcomp", "ccomp", "xcomp"]}, "POS": {"IN": ["VERB", "ADV", "ADP"]}}
        },
    ],
    "num_quantmod_chain" : [ # just under 296, at least 74
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM"}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "quant",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "amod", "compound", "nummod", "nmod", "advmod", "npadvmod"]},
                            "POS": {"IN": ["ADJ", "SCONJ", "ADV", "PART", "ADP", "DET"]}}
        },
        {
            "LEFT_ID": "quant",
            "REL_OP": ">",
            "RIGHT_ID": "quant2",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "amod", "compound", "nummod", "nmod", "advmod", "npadvmod"]},
                            "POS": {"IN": ["ADJ", "SCONJ", "ADV", "PART", "ADP", "DET"]}} # ADDED DET
        },
    ],

    "verb_quantmod_num" : [ # fell by 25
        {
            "RIGHT_ID": "num",
            "RIGHT_ATTRS": {"POS": "NUM"}
        },
        {
            "LEFT_ID": "num",
            "REL_OP": ";",
            "RIGHT_ID": "quant",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["by"]}}
        },
        {
            "LEFT_ID": "quant",
            "REL_OP": "<",
            "RIGHT_ID": "verb",
            "RIGHT_ATTRS": {"POS": {"IN": ["VERB"]}}
        }
    ],

    "verb_quantmod2_num" : [ # evised upward by Y=40 billion
        {
            "RIGHT_ID": "num",
            "RIGHT_ATTRS": {"POS": "NUM"}
        },
        {
            "LEFT_ID": "num",
            "REL_OP": ";",
            "RIGHT_ID": "quant",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["by"]}}
        },
        {
            "LEFT_ID": "quant",
            "REL_OP": "<",
            "RIGHT_ID": "verb",
            "RIGHT_ATTRS": {"POS": {"IN": ["VERB"]}}
        },
        {
            "LEFT_ID": "verb",
            "REL_OP": ">",
            "RIGHT_ID": "quant2",
            "RIGHT_ATTRS": {"DEP": {"IN": ["advmod"]},"POS": {"IN": ["ADV"]}}
        }
    ],

    "num_right_noun" : [ # $62 and $68 per share (62 share; 68 share -> separate matches)
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM"}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">>",
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"DEP": {"IN": ["prep", "pobj"]}, "POS": {"IN": ["PROPN", "NOUN"]}}
        },
    ],

    # If it behaves like a number, its probably a number
    "noun_noun" : [ # tens of thousands, 75 kwW system, 68 league games, 50 mph winds
        # but also state and local tax credits, iPhone Mini deal, most cloud-to-ground
        {
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "nummod", "nmod","compound"]}, "POS": {"IN": ["PROPN", "NOUN"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "quantifier",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "amod", "nummod", "nmod", "advmod", "npadvmod"]}}
        },
    ],

    "noun_noun_2" : [ # 40 yard line
        {
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "nummod", "nmod", "compound", "npadvmod"]}, "POS": {"IN": ["PROPN", "NOUN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "quantifier",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "nummod", "nmod", "advmod", "npadvmod"]}}
        },
    ],

    "num_num" : [ # 7.2 billion, 13.6 million, 1 trillion
        # 1.0 l, 2.7 m, 7 s
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM", "_": {"like_scale": True}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"DEP": {"IN": ["compound", "nummod"]}, "POS": {"IN": ["NUM"]}}
        },
    ],

    "num_to_num" : [ # 40 to 50, 100 to 200
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM"}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ".",
            "RIGHT_ID": "part",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "prep"]}, "POS": {"IN":["PART", "ADP"]}, "ORTH": {"IN": ["to"]}}
        },
        {
            "LEFT_ID": "part",
            "REL_OP": ".",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        }
    ],

    "num_to_num_2" : [ # $920 to $1730 a week
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM", "IS_DIGIT": True}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ".",
            "RIGHT_ID": "part",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "prep"]}, "POS": {"IN":["PART", "ADP"]}, "ORTH": {"IN": ["to"]}}
        },
        {
            "LEFT_ID": "part",
            "REL_OP": ">",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        }
    ],

    "num_to_num_3" : [ # 32 million to 33 million, 15 billion to 16 billion euros
        {
            "RIGHT_ID": "number", # 32
            "RIGHT_ATTRS": {"POS": "NUM", "IS_DIGIT": True}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": "<",
            "RIGHT_ID": "number2", # million
            "RIGHT_ATTRS": {"POS": "NUM", "_": {"like_scale": True}}
        },
        {
            "LEFT_ID": "number2",
            "REL_OP": ".",
            "RIGHT_ID": "part", # to
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "prep"]}, "POS": "PART", "ORTH": {"IN": ["to"]}}
        },
        {
            "LEFT_ID": "part",
            "REL_OP": "<",
            "RIGHT_ID": "number3", # million
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}, "_": {"like_scale": True}}
        },
        {
            "LEFT_ID": "number3",
            "REL_OP": ">",
            "RIGHT_ID": "number4", # 33
            "RIGHT_ATTRS": {"POS": "NUM", "IS_DIGIT": True}
        },
    ],

    "num_to_num_4" : [ # 90 billion to 100 billion
        {
            "RIGHT_ID": "number", # billion
            "RIGHT_ATTRS": {"POS": "NUM",  "_": {"like_scale": True}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "number2", # 90
            "RIGHT_ATTRS": {"POS": "NUM", "IS_DIGIT": True}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "number3", # 100
            "RIGHT_ATTRS": {"POS": "NUM", "IS_DIGIT": True}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "number4", # billion
            "RIGHT_ATTRS": {"POS": "NUM", "_": {"like_scale": True}}
        },
        {
            "LEFT_ID": "number4",
            "REL_OP": ".",
            "RIGHT_ID": "part", # to
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "prep"]}, "POS": "PART", "ORTH": {"IN": ["to"]}}
        }
    ],

    "num_to_num_5" : [ # $48 to $55
        {
            "RIGHT_ID": "number", # 55
            "RIGHT_ATTRS": {"POS": "NUM"}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "part", # to
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "prep"]}, "POS": {"IN":["PART", "ADP"]}, "ORTH": {"IN": ["to"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "number2", # 48
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        }
    ],

    "num_to_num_dig" : [ # 5-10
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM", "IS_DIGIT": True}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ".",
            "RIGHT_ID": "part",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["-"]}}
        },
        {
            "LEFT_ID": "part",
            "REL_OP": ".",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}, "IS_DIGIT": True}
        },

    ],

    "num_to_num_num" : [ # comment out
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM", "_": {"like_scale": False}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "part",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["to"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod"]}, "POS": {"IN": ["NUM"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "number_compound",
            "RIGHT_ATTRS": {"DEP": {"IN": ["compound"]}, "POS": {"IN": ["NUM"]}}
        },

    ],

    "num_to_num_num_dig" : [ # comment out
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM", "IS_DIGIT": True}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "part",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["-"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod"]}, "POS": {"IN": ["NUM"]}, "IS_DIGIT": True}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "number_compound",
            "RIGHT_ATTRS": {"DEP": {"IN": ["compound"]}, "POS": {"IN": ["NUM"]}}
        },

    ],

    "range_single" : [ # thousands of students/dollars/deals, millions of Americans, dozens of people/vehicles
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NOUN", "_": {"like_number": True}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "of",
            "RIGHT_ATTRS": {"DEP": {"IN": ["prep"]}, "ORTH": {"IN": ["of"]}}
        },
        {
            "LEFT_ID": "of",
            "REL_OP": ">",
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"DEP": {"IN": ["pobj"]}, "POS": {"IN": ["NOUN", "PROPN"]}}
        },

    ],
    "range_double" : [
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NOUN", "_": {"like_number": True}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "of",
            "RIGHT_ATTRS": {"DEP": {"IN": ["prep"]}, "ORTH": {"IN": ["of"]}}
        },
        {
            "LEFT_ID": "of",
            "REL_OP": ">",
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"DEP": {"IN": ["pobj"]}, "POS": {"IN": ["NOUN"]}, "_": {"like_number": True}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "of2",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod"]}, "ORTH": {"IN": ["of"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod"]}, "POS": {"IN": ["NOUN"]}}
        },
    ],
    "frac" : [ # one of four
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM"}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ".",
            "RIGHT_ID": "of",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "prep"]}, "ORTH": {"IN": ["of"]}}
        },
        {
            "LEFT_ID": "of",
            "REL_OP": ".",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"DEP": {"IN": ["pobj", "appos", "nummod"]}, "POS": {"IN": ["NUM"]}}
        },
    ],

    "frac_2" : [ # three out of five, one out of three
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM"}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ".",
            "RIGHT_ID": "out",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["out"]}}
        },
        {
            "LEFT_ID": "out",
            "REL_OP": ".",
            "RIGHT_ID": "of",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["of"]}}
        },
        {
            "LEFT_ID": "of",
            "REL_OP": ".",
            "RIGHT_ID": "number_2",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        }
    ],

    "unit_frac" : [ # 68 per share, miles per hour, million per worker
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM", "NOUN"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "per",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["per", "a", "an"]}}
        },
        {
            "LEFT_ID": "per",
            "REL_OP": ">",
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["NOUN", "PROPN"]}}
        },
    ],

    "unit_frac_2" : [ # 14.95 a month, 1730 a week
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["NOUN", "PROPN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "per",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["per", "a", "an"]}}
        },
    ],

    "unit_frac_3" : [ # 21.25 per square foot
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "per",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["per"]}}
        },
        {
            "LEFT_ID": "per",
            "REL_OP": ">",
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["NOUN", "PROPN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "adj",
            "RIGHT_ATTRS": {"POS": {"IN": ["ADJ"]}}
        },
    ],
    "unit_frac_4" : [ # parts per million, tests per day, miles per hour
        {
            "RIGHT_ID": "noun", # parts
            "RIGHT_ATTRS": {"POS": {"IN": ["NOUN", "PROPN", "SYM"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "per", # per
            "RIGHT_ATTRS": {"DEP": {"IN": ["prep"]}, "ORTH":{"IN": ["per", "a", "an"]}}
        },
        {
            "LEFT_ID": "per",
            "REL_OP": ">",
            "RIGHT_ID": "number2", # million
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM", "NOUN", "PROPN"]}}
        },
    ],

    "unit_frac_5" : [ # 9 billion shekels ($2.408 billion) a year, RM a month, barrels a day
        {
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["NOUN", "PROPN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "appos",
            "RIGHT_ATTRS": {"ORTH":{"IN": ["year", "month", "day", "hour", "minute", "second"]}}
        },
        {
            "LEFT_ID": "appos",
            "REL_OP": ">",
            "RIGHT_ID": "per",
            "RIGHT_ATTRS": {"POS": {"IN": ["DET"]}, "ORTH":{"IN": ["per", "a", "an"]}}
        },
    ],

    "unit_frac_6" : [ # 50 hours a week
        {
            "RIGHT_ID": "noun", # hours
            "RIGHT_ATTRS": {"POS": {"IN": ["NOUN", "PROPN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "noun2", # week
            "RIGHT_ATTRS": {"POS": {"IN": ["NOUN", "PROPN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ".",
            "RIGHT_ID": "per", # a
            "RIGHT_ATTRS": {"ORTH":{"IN": ["per", "a", "an"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ";",
            "RIGHT_ID": "number", # 50
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        }
    ],

    "dimensions_3" : [
        {
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "nmod", "nummod"]}, "POS": "NUM"}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": "$+",
            "RIGHT_ID": "punkt",
            "RIGHT_ATTRS": {"POS": "SYM"}
        },
        {
            "LEFT_ID": "punkt",
            "REL_OP": "$+",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "nmod", "nummod"]}, "POS": "NUM"}
        },
        {
            "LEFT_ID": "number2",
            "REL_OP": "$+",
            "RIGHT_ID": "punkt2",
            "RIGHT_ATTRS": {"POS": "SYM"}
        },

        {
            "LEFT_ID": "punkt2",
            "REL_OP": "$+",
            "RIGHT_ID": "number3",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "nmod", "nummod"]}, "POS": "NUM"}
        }
    ],

    "dimensions_2" : [
        {
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "nmod", "nummod"]}, "POS": "NUM"}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": "$+",
            "RIGHT_ID": "punkt",
            "RIGHT_ATTRS": {"POS": "SYM"}
        },
        {
            "LEFT_ID": "punkt",
            "REL_OP": "$+",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod", "nmod", "nummod"]}, "POS": "NUM"}
        }

    ],

    "noun_quant_noun_noun" : [ # 256gb screeen size, 50 foot building lots, ... billion dalkon shield claimants, ... million world bank loan 
        {
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"DEP": "nummod", "POS": "NUM"}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "noun2",
            "RIGHT_ATTRS": {"DEP":"compound","POS":"NOUN" }
        },
        {
            "LEFT_ID": "noun2",
            "REL_OP": ">",
            "RIGHT_ID": "noun3",
            "RIGHT_ATTRS": {"DEP":"compound","POS":"NOUN" }
        }
    ],

    "adp_num_cconj_num" : [ # between 50 and 70, from 13.5 to 14.0
        {
            "RIGHT_ID": "adp",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["between", "from"]}}
        },
        {
            "LEFT_ID": "adp",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "cconj",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["and", "to"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        },
    ],

    "adp_num_cconj_num_2" : [ # from 5.7% to 3.4%, from 2 dollar to 3 dollar
        # from (up to) 252 miles to 287 miles-
        {
            "RIGHT_ID": "adp",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["from"]}}
        },
        {
            "LEFT_ID": "adp",
            "REL_OP": ">",
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["NOUN", "SYM"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        },
        {
            "LEFT_ID": "adp",
            "REL_OP": ">",
            "RIGHT_ID": "adp2",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["to"]}}
        },
        {
            "LEFT_ID": "adp2",
            "REL_OP": ">",
            "RIGHT_ID": "noun2",
            "RIGHT_ATTRS": {"DEP": {"IN": ["pobj"]}}
        },
        {
            "LEFT_ID": "noun2",
            "REL_OP": ">",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        },
    ],

    "adp_num_cconj_num_3" : [ # Fresh fruit prices are expected to increase between 5% and 6%, with dairy prices expected to increase between 4% and 5% and fats and oils between 6% and 7%.
        {
            "RIGHT_ID": "noun2",
            "RIGHT_ATTRS": {"POS": {"IN": ["NOUN"]}}
        },
        {
            "LEFT_ID": "noun2",
            "REL_OP": ">",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        },
        {
            "LEFT_ID": "noun2",
            "REL_OP": "<",
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["NOUN"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "adp2",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["and"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": "<",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "adp",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["between"]}}
        }
    ],

    "adp_num_cconj_num_with_scale" : [ # between 60m and 90m subscribers
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"POS": "NUM"}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "adp",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["and"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "scale",
            "RIGHT_ATTRS": {"DEP": {"IN": ["quantmod"]}, "POS": {"IN": ["NOUN"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": "<",
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN", "SYM"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": "<",
            "RIGHT_ID": "adp2",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["between"]}}
        }
    ],

    "adp_num_cconj_num_with_unit" : [ # between $62 and $68 per share, between 4 and 5
        {
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}, "ORTH": {"NOT_IN": [ "thousand", "million", "billion", "trillion"]}, "LIKE_NUM": True}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}, "LIKE_NUM": True}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "adp",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["between"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "adp2",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["and"]}}
        }
    ],

    "adp_num_cconj_num_with_unit_2" : [ # between 20 HZ and 20,000 HZ
        {
            "RIGHT_ID": "number", # 20
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "adp", # between
            "RIGHT_ATTRS": {"ORTH": {"IN": ["between"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": "<",
            "RIGHT_ID": "noun", # HZ
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN", "SYM"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "adp2", # and
            "RIGHT_ATTRS": {"ORTH": {"IN": ["and"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "noun2", # HZ
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN", "SYM"]}}
        }
    ],

    "adp_num_cconj_num_with_unit_3" : [ # from 2,415 to 2,315, from 160 to 200
        {
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"POS": "NUM", "_": {"like_number": True}}
        },
        {
            "LEFT_ID": "number2",
            "REL_OP": "<",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": "NUM", "_": {"like_number": True}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": "<",
            "RIGHT_ID": "adp",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["from"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "adp2",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["to"]}}
        }
    ],

    "adp_num_cconj_num_with_unit_4" : [ # between 15kHz and 17
        {
            "RIGHT_ID": "adp",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["between"]}}
        },
        {
            "LEFT_ID": "adp",
            "REL_OP": ">",
            "RIGHT_ID": "number1",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}, "_": {"like_number": True}}
        },
        {
            "LEFT_ID": "adp",
            "REL_OP": ">",
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["PROPN", "NOUN", "SYM"]}}
        },
        {
            "LEFT_ID": "noun",
            "REL_OP": ">",
            "RIGHT_ID": "adp2",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["and"]}}
        },
        {
            "LEFT_ID": "adp2",
            "REL_OP": ".",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}, "_": {"like_number": True}}
        }
    ],

    "adp_num_cconj_num_with_unit_5" : [ #between US$ 160 million and US$ 171 million, between 160 million and 171
        {
            "RIGHT_ID": "number", # million
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}, "_": {"like_scale": True}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "number2", # 160
            "RIGHT_ATTRS": {"POS": "NUM", "IS_DIGIT": True}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "adp",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["between"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "adp2",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["and"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "number3", # million
            "RIGHT_ATTRS": {"POS": "NUM", "_": {"like_scale": True}}
        },
        {
            "LEFT_ID": "number3",
            "REL_OP": ">",
            "RIGHT_ID": "number4", # 171
            "RIGHT_ATTRS": {"POS": "NUM", "IS_DIGIT": True}
        },
    ],

    "adp_num_cconj_num_with_unit_6" : [ # between ¥ 60 and ¥ 110, between 5% and 6%
        {
            "RIGHT_ID": "adp",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["between"]}}
        },
        {
            "LEFT_ID": "adp",
            "REL_OP": "<",
            "RIGHT_ID": "number",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "adp2",
            "RIGHT_ATTRS": {"ORTH": {"IN": ["and"]}}
        },
        {
            "LEFT_ID": "number",
            "REL_OP": ">",
            "RIGHT_ID": "sym",
            "RIGHT_ATTRS": {"POS": {"IN": ["SYM", "NOUN"]}}
        },
        {
            "LEFT_ID": "sym",
            "REL_OP": ">",
            "RIGHT_ID": "number2",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}}
        }
    ],

    "compound_num": [ # five hundred
        {
            "RIGHT_ID": "num",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}, "IS_DIGIT": False}
        },
        {
            "LEFT_ID": "num",
            "REL_OP": ">",
            "RIGHT_ID": "num2",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"], "DEP": {"IN": ["nummod", "compound", "quantmod"]}}, "IS_DIGIT": False}
        },
    ],
    "compound_num_2": [ # [thousand, million, three] from "three million five hundred thousand"
        {
            "RIGHT_ID": "num",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"]}, "IS_DIGIT": False}
        },
        {
            "LEFT_ID": "num",
            "REL_OP": ">",
            "RIGHT_ID": "num2",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"], "DEP": {"IN": ["nummod", "compound", "quantmod"]}}, "IS_DIGIT": False}
        },
        {
            "LEFT_ID": "num2",
            "REL_OP": ">",
            "RIGHT_ID": "num3",
            "RIGHT_ATTRS": {"POS": {"IN": ["NUM"], "DEP": {"IN": ["nummod", "compound", "quantmod"]}}, "IS_DIGIT": False}
        },
    ],

    "one_of": [ # one of the computers, one of the proudest moments in his life
        {
            "RIGHT_ID": "num",
            "RIGHT_ATTRS": {"POS": "NUM"}
        },
        {
            "LEFT_ID": "num",
            "REL_OP": ">",
            "RIGHT_ID": "of",
            "RIGHT_ATTRS": {"POS": {"IN": ["ADP"]}, "ORTH": {"IN": ["of"]}}
        },
        {
            "LEFT_ID": "of",
            "REL_OP": ">",
            "RIGHT_ID": "noun",
            "RIGHT_ATTRS": {"POS": {"IN": ["NOUN"]}}
        }
    ],

    "phone_number_pattern_1" : [ # (0) 20 111 2222
        {"ORTH": "("},
        {"SHAPE": "d"},
        {"ORTH": ")"},
        {"SHAPE": "dd"},
        {"ORTH": "-", "OP": "?"},
        {"SHAPE": "ddd"},
        {"ORTH": "-", "OP": "?"},
        {"SHAPE": "dddd"}
    ],

    "phone_number_pattern_2" : [ # (0)20 111 2222
        {"ORTH": "("},
        {"TEXT": {"REGEX": r"[\d]\)[\d]*"}},
        {"ORTH": "-", "OP": "?"},
        {"SHAPE": "ddd"},
        {"ORTH": "-", "OP": "?"},
        {"SHAPE": "dddd"}
    ],

    "phone_number_pattern_3" : [ # (123) 4567 8901, (123) 4567-8901
        {"ORTH": "("},
        {"SHAPE": "ddd"},
        {"ORTH": ")"},
        {"SHAPE": "dddd"},
        {"ORTH": "-", "OP": "?"},
        {"SHAPE": "dddd"}
    ],

    "phone_number_pattern_4" : [ # (123) 456 789
        {"ORTH": "("},
        {"SHAPE": "ddd"},
        {"ORTH": ")"},
        {"SHAPE": "ddd"},
        {"ORTH": "-", "OP": "?"},
        {"SHAPE": "ddd"}
    ],

    "phone_number_pattern_5" : [ # + AA (AAAA) BBBBBBB (international numbers)
        {"ORTH": "+"},
        {"SHAPE": "dd"},
        {"ORTH": "(", "OP": "?"},
        {"SHAPE": "dddd"},
        {"ORTH": ")", "OP": "?"},
        {"TEXT": {"REGEX": r"\d{4,7}"}}
    ],

    "zip_number_pattern" : [ # 90049 ZIP Code
        {"TEXT": {"REGEX": r"\d{5}"}},
        {"ORTH": "ZIP"}
    ],

    #
    #     "noun_quant_punkt" : [#~10tw/h
    #         {
    #             "RIGHT_ID": "adp",
    #             "RIGHT_ATTRS": {"POS":"ADP"},
    #         },
    #
    #         {
    #             "RIGHT_ID": "noun",
    #             "RIGHT_ATTRS": {"DEP": "pobj"},
    #             "LEFT_ID": "adp",
    #             "REL_OP": ";*"}
    # ]
}



from CQE.unit_classifier.train_classifier_bert import ambigious_units
from pathlib import Path

TOPDIR = Path(__file__).parent.parent.parent
import spacy
import operator
import os
def get_project_root() -> Path:
    return Path(__file__).parent.parent
#######
#contains code for the disambiguator class, which is combination of bert classifiers
#given a surface form the class returns a correct classifer output using the .cats from the spacy transformer
#models should be located in data/units/unit_models/ for this class to work correctly 
######
class unit_disambiguator():
    def __init__(self):
        self.models = {}
        path = get_project_root()
        for key, values in ambigious_units.items():
            if key in ["C", "B", "P"]:
                file_name = os.path.join(path, "unit_models/train_BIG" + key + ".spacy/model-best")
            elif key=="¥":
                file_name = os.path.join(path, "unit_models/train_yen.spacy/model-best")
            elif key=="′":
                file_name = os.path.join(path, "unit_models/train_ascii'.spacy/model-best")
            elif key=="″":
                file_name = os.path.join(path, "unit_models/train_ascii_doublequote.spacy/model-best")
            elif key=="\"":
                file_name = os.path.join(path, "unit_models/train_doublequote.spacy/model-best")

            else:
                file_name = os.path.join(path, "unit_models/train_" + key + ".spacy/model-best")
            self.models[key] = spacy.load(file_name)

    def disambiguate(self, sentence, surface_form):
        probabilities = self.models[surface_form](sentence).cats
        return max(probabilities.items(), key=operator.itemgetter(1))[0]



from unit_disambiguator import unit_disambiguator


dis=unit_disambiguator()
print(dis.disambiguate("I ate so much and I have gained 10 pounds.","pound"))
print(dis.disambiguate("he spend around 20 pounds.","pound"))
print(dis.disambiguate("john is going to buy a new play station for around 500 pounds","pound"))
print(dis.disambiguate("this couch is heavy, maybe 20 pounds","pound"))
print(dis.disambiguate("the temperature is 2c","c"))
print(dis.disambiguate("nasdaq rose by 2p","p"))
print(dis.disambiguate("each pot of land is land measuring 0.5 a","a"))


from setuptools import setup
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='CQE',
    version='1.0.6',
    packages=['CQE','CQE.unit_classifier'],
    package_data={'': ['unit.json'],'': ['unit_models.zip']},# both has to be empty
    url='https://github.com/vivkaz/CQE',
    license='',
    long_description_content_type="text/markdown",
    long_description=long_description,
    author='satyaalmasian and vivian kazakova',
    author_email='satya.almasian@gmail.com',
    description='quantity extractor',
    install_requires=['fuzzywuzzy==0.18.0', 'more_itertools','ordered-set','python-Levenshtein',
                      'spacy==3.0.9', 'greek','spacy-legacy','requests',
                      'regex','emoji==1.7','torch==2.0.0','spacy-transformers==1.0.4','protobuf==3.20.1','inflect==5.4.0',
                      'spacy_download'],
)









import json
import os
import re
from pathlib import Path
from decimal import Decimal
from typing import Iterable
from fuzzywuzzy import fuzz
from fuzzywuzzy import process

def get_project_root() -> Path:
    return Path(__file__).parent


class Change:
    def __init__(self, change="", span=[]):
        self.span = span
        self.change = change

    def __str__(self):
        return self.change

    def __bool__(self):
        return bool(self.change)


class Range:
    def __init__(self, lower, upper, span=[], overload=False):
        self.span = sorted(span, key=lambda t: t.i)
        self.lower = lower
        self.upper = upper
        self.char_indices = [(token.idx, token.idx + len(token.text)) for token in span if token.pos_ in ["NUM", "NOUN"]] if span and overload else None
        self.scientific_notation_lower = '%e' % Decimal(str(self.lower)) # e.g. 0.000000e+00
        self.scientific_notation_upper = '%e' % Decimal(str(self.upper)) # e.g. 8.000000e+01
        self.scientific_notation = self.scientific_notation_lower + "-" + self.scientific_notation_upper
        self.simplified_scientific_notation_lower, self.simplified_scientific_notation_upper = self.get_simplified_scientific_notation() # e.g. 0e+00, 8e+01
        self.simplified_scientific_notation = self.simplified_scientific_notation_lower + "-" + self.simplified_scientific_notation_upper

    def __str__(self):
        return f"{self.lower}-{self.upper}"

    def __bool__(self):
        return bool(self.lower is not None and self.upper is not None)

    def get_str_value(self, idx_value=0):
        if len(self.char_indices) == 1: # e.g. dozens, millions, thousands etc.
            lower = str(int(self.lower)) if self.lower - int(self.lower) == 0 else str(self.lower)
            upper = str(int(self.upper)) if self.upper - int(self.upper) == 0 else str(self.upper)
            return f"{lower}-{upper}"
        if idx_value == 0:
            return str(int(self.lower)) if self.lower - int(self.lower) == 0 else str(self.lower)
        return str(int(self.upper)) if self.upper - int(self.upper) == 0 else str(self.upper)

    def get_simplified_scientific_notation(self):
        """return a simplified scientific notation of the lower and upper value"""
        notation_lower = '%e' % Decimal(str(self.lower))
        notation_upper = '%e' % Decimal(str(self.upper))
        return notation_lower.split('e')[0].rstrip('0').rstrip('.') + 'e' + notation_lower.split('e')[1], notation_upper.split('e')[0].rstrip('0').rstrip('.') + 'e' + notation_upper.split('e')[1]


class Value:
    def __init__(self, value, span=[], overload=False):
        self.span = sorted(span, key=lambda t: t.i)
        self.value = value
        #self.char_indices = [(token.idx, token.idx + len(token.text)) for token in span] if span and overload else None
        self.char_indices = [(span[0].idx, span[-1].idx + len(span[-1].text))] if span and overload else None
        self.scientific_notation = '%e' % Decimal(self.__str__()) # e.g. 4.000000e-03
        self.simplified_scientific_notation = self.get_simplified_scientific_notation() # e.g. 4e-03

    def __str__(self):
        return str(self.value)

    def __bool__(self):
        return bool(self.value is not None)

    def get_str_value(self, idx_value=0):
        if self.value - int(self.value) == 0:
            return str(int(self.value))
        return str(self.value)

    def get_simplified_scientific_notation(self):
        """return a simplified scientific notation of the value"""
        notation = '%e' % Decimal(self.__str__())
        return notation.split('e')[0].rstrip('0').rstrip('.') + 'e' + notation.split('e')[1]


class Unit:
    def __init__(self, unit=[], norm_unit="-", span=[], scientific=False, keys=[], overload=False):
        self.span = sorted(span, key=lambda t: t.i) if norm_unit != "-" else []
        if len(set(token.text for token in unit)) != 1 and len(unit) > 1:
            self.unit = [token for i,token in enumerate(unit) if token.text != unit[(i+1)%len(unit)].text]
        else:
            self.unit = unit
        self.norm_unit = norm_unit
        if len(span)>1 and sorted([token.i for token in span]) == list(range(min(token.i for token in span), max(token.i for token in span)+1)):
            self.char_indices = [(span[0].idx, span[-1].idx + len(span[-1].text))] if span and overload else None
        else:
            self.char_indices = [(token.idx, token.idx + len(token.text)) for token in span] if span and overload else None
        self.scientific = scientific if overload else None # unit type attribute, which identifies if it is a noun based unit, or a scientific unit (we find it in out dict)
        self.unit_keys = self.set_list_of_keys(keys) if overload else None # list of (lists of) keys from the unit dict used for normalization
        self.unit_surfaces_forms = self.set_unit_surface_forms() if overload else None # all the surface forms of the unit that is detected

    def __str__(self):
        return str(self.unit)

    def __bool__(self):
        return bool(self.unit)
    
    def _flatten(self, items):
        """yield items from nested iterable"""
        for item in items:
            if isinstance(item, Iterable) and not isinstance(item, (str, bytes)):
                for sub_item in self._flatten(item):
                    yield sub_item
            else:
                yield item

    def set_list_of_keys(self, keys_list):
        """return a list of the keys used for the normalization of the unit"""
        keys = []
        k_list = list(self._flatten(keys_list))
        for key in k_list:
            keys.append(key)
        return keys

    def set_unit_surface_forms(self):
        """return list or dict of the surface forms for the unit"""
        if self.scientific and self.unit_keys: # in unit dictionary
            path = get_project_root()
            file_name = os.path.join(path, "unit.json")
            with open(file_name, "r", encoding="utf8") as f:
                units_dict = json.load(f)
            if len(self.unit_keys) >= 1: # compound unit e.g. 'dollar per square foot' or single unit e.g. 'percentage'
                return { key: units_dict[key].get("surfaces")+units_dict[key].get("symbols") for key in self.unit_keys }
        return {} # noun based unit => no surface forms


class Concept:
    def __init__(self, noun=[], span=[]):
        self.span = sorted(span, key=lambda t: t.i)
        if noun:
            self.noun = {}
            for tokens_list in noun:
                if tokens_list: # consider only not empty lists
                    self.noun.update({len(self.noun): tokens_list})
            if not self.noun: # empty => "-"
                self.noun = "-"
        else:
            self.noun = "-" # no noun found => "-"

    def __str__(self):
        return str(self.noun)

    def __bool__(self):
        return bool(self.noun)

    def get_nouns(self):
        """return the referred nouns from the dict in a list form"""
        return [list(list_noun) for list_noun in self.noun.values()] if self.noun != "-" else []


class Quantity:
    original_text = None # class variable

    def __init__(self, value, change=Change(), unit=Unit(), referred_concepts=None, in_bracket=False):
        self.change = change
        self.value = value
        self.unit = unit
        self.referred_concepts = [] if not referred_concepts else referred_concepts
        self.__value_char_indices = value.char_indices # characters indices based on the preprocessed sentence
        self.__unit_char_indices = unit.char_indices # characters indices based on the preprocessed sentence
        self.__original_value_char_indices = None # characters indices based on the original sentence
        self.__original_unit_char_indices = None # characters indices based on the original sentence
        self.__preprocessed_text = None # default
        self.__normalized_text = None # default
        self.in_bracket = in_bracket

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"({str(self.change)},{str(self.value)},{str(self.unit.unit)},{str(self.unit.norm_unit)},{str(self.referred_concepts)})"
    
    def set_original_text(self, text):
        Quantity.original_text = text

    def set_preprocessed_text(self, text):
        self.__preprocessed_text = text

    def set_normalized_text(self, text):
        self.__normalized_text = text

    def get_preprocessed_text(self):
        return self.__preprocessed_text
    
    def get_normalized_text(self):
        return self.__normalized_text

    def transform_concept_to_dict(self):
        """create a Concept instance"""
        self.referred_concepts = Concept(self.referred_concepts)

    def get_unit_surface_forms(self):
        """return list or dict of the surface forms for the Unit"""
        return self.unit.unit_surfaces_forms
    
    def get_char_indices(self):
        """return dictionary of the Value and Unit indices in the preprocessed sentence"""
        return { "value": self.__value_char_indices, "unit": self.__unit_char_indices }
    
    def compute_original_char_indices(self):
        """return characters indices based on the the original sentence"""

        def find_most_similar_substring(target_substring, text):
            """find the most similar substring to the given target within the text"""

            def n_grams(tokens, n):
                """generate n-grams from a list of tokens"""
                n_grams = []
                for j in range(1,n+1):
                    n_grams.extend(["".join(tokens[i:i+j]) for i in range(len(tokens)-j+1)])
                return n_grams

            return process.extractOne(target_substring, n_grams(text, len(text)), scorer=fuzz.ratio)

        # value indices
        original_value_indices = []
        value_indices = self.__value_char_indices
        target = self.__preprocessed_text[sorted(value_indices)[0][0]:sorted(value_indices)[-1][1]]
        
        # find matches of the target substring in the original text
        matches = list(re.finditer(re.escape(target), self.original_text))

        if not matches:
            # if no direct matches, find the most similar substring and find matches
            most_similar = find_most_similar_substring(target, self.original_text)[0]
            matches = re.finditer(re.escape(most_similar), self.original_text)

        # iterate through the matches and mark already observed ones
        for match in matches:
            original_value_indices.append((match.start(), match.end()))
            mark = "@"*len(match.group())
            self.set_original_text(re.sub(f"\b{re.escape(match.group())}\b", mark, self.original_text, count=1))
            break

        # unit indices
        original_unit_indices = []
        unit_indices = self.__unit_char_indices if self.__unit_char_indices else []
        target = ""

        # iterate through each span in the sorted list of spans
        for tuple_idx in sorted(unit_indices):
            target = self.__preprocessed_text[tuple_idx[0]:tuple_idx[1]]
            
            # find matches of the target substring in the original text
            matches = list(re.finditer(re.escape(target), self.original_text))

            if not matches:
                # if no direct matches, find the most similar substring and find matches
                most_similar = find_most_similar_substring(target, self.original_text)[0]
                matches = re.finditer(re.escape(most_similar), self.original_text)

            # iterate through the matches and mark already observed ones
            for match in matches:
                original_unit_indices.append((match.start(), match.end()))
                mark = "@"*len(match.group())
                self.set_original_text(re.sub(f"\b{re.escape(match.group())}\b", mark, self.original_text, count=1))
                break

        return original_value_indices, original_unit_indices


    def get_original_char_indices(self):
        """return dictionary of the Value and Unit indices in the original sentence"""
        if self.__original_value_char_indices is None and self.__original_unit_char_indices is None:
            self.__original_value_char_indices, self.__original_unit_char_indices = self.compute_original_char_indices()

        return { "value": self.__original_value_char_indices, "unit": self.__original_unit_char_indices }
